import { StreamClient, APIResponse, UnknownRecord } from './client';
declare type BaseCollection<CollectionType> = {
    collection: string;
    data: CollectionType;
    id: string;
};
export declare type CollectionResponse<CollectionType extends UnknownRecord = UnknownRecord> = BaseCollection<CollectionType> & {
    created_at: string;
    foreign_id: string;
    updated_at: string;
};
export declare type NewCollectionEntry<CollectionType extends UnknownRecord = UnknownRecord> = BaseCollection<CollectionType> & {
    user_id?: string;
};
export declare type CollectionAPIResponse<CollectionType extends UnknownRecord = UnknownRecord> = APIResponse & CollectionResponse<CollectionType>;
export declare type SelectCollectionAPIResponse<CollectionType extends UnknownRecord = UnknownRecord> = APIResponse & {
    response: {
        data: CollectionResponse<CollectionType>[];
    };
};
export declare type UpsertCollectionAPIResponse<CollectionType extends UnknownRecord = UnknownRecord> = APIResponse & {
    data: {
        [key: string]: {
            data: CollectionType;
            id: string;
        }[];
    };
};
export declare class CollectionEntry<CollectionType extends UnknownRecord = UnknownRecord> {
    id: string;
    collection: string;
    store: Collections<CollectionType>;
    data: CollectionType | null;
    full?: unknown;
    constructor(store: Collections<CollectionType>, collection: string, id: string, data: CollectionType);
    ref(): string;
    /**
     * get item from collection and sync data
     * @method get
     * @memberof CollectionEntry.prototype
     * @return {Promise<CollectionEntry<CollectionType>>}
     * @example collection.get("0c7db91c-67f9-11e8-bcd9-fe00a9219401")
     */
    get(): Promise<CollectionEntry<CollectionType>>;
    /**
     * Add item to collection
     * @link https://getstream.io/docs/collections_introduction/?language=js#adding-collections
     * @method add
     * @memberof CollectionEntry.prototype
     * @return {Promise<CollectionEntry<CollectionType>>}
     * @example collection.add("cheese101", {"name": "cheese burger","toppings": "cheese"})
     */
    add(): Promise<CollectionEntry<CollectionType>>;
    /**
     * Update item in the object storage
     * @link https://getstream.io/docs/collections_introduction/?language=js#updating-collections
     * @method update
     * @memberof CollectionEntry.prototype
     * @return {Promise<CollectionEntry<CollectionType>>}
     * @example store.update("0c7db91c-67f9-11e8-bcd9-fe00a9219401", {"name": "cheese burger","toppings": "cheese"})
     * @example store.update("cheese101", {"name": "cheese burger","toppings": "cheese"})
     */
    update(): Promise<CollectionEntry<CollectionType>>;
    /**
     * Delete item from collection
     * @link https://getstream.io/docs/collections_introduction/?language=js#removing-collections
     * @method delete
     * @memberof CollectionEntry.prototype
     * @return {Promise<APIResponse>}
     * @example collection.delete("cheese101")
     */
    delete(): Promise<APIResponse>;
}
export declare class Collections<CollectionType extends UnknownRecord = UnknownRecord> {
    client: StreamClient;
    token: string;
    /**
     * Initialize a feed object
     * @method constructor
     * @memberof Collections.prototype
     * @param {StreamCloudClient} client Stream client this collection is constructed from
     * @param {string} token JWT token
     */
    constructor(client: StreamClient, token: string);
    buildURL: (collection: string, itemId?: string | undefined) => string;
    entry(collection: string, itemId: string, itemData: CollectionType): CollectionEntry<CollectionType>;
    /**
     * get item from collection
     * @link https://getstream.io/docs/collections_introduction/?language=js#retrieving-collections
     * @method get
     * @memberof Collections.prototype
     * @param  {string}   collection  collection name
     * @param  {string}   itemId  id for this entry
     * @return {Promise<CollectionEntry<CollectionType>>}
     * @example collection.get("food", "0c7db91c-67f9-11e8-bcd9-fe00a9219401")
     */
    get(collection: string, itemId: string): Promise<CollectionEntry<CollectionType>>;
    /**
     * Add item to collection
     * @link https://getstream.io/docs/collections_introduction/?language=js#adding-collections
     * @method add
     * @memberof Collections.prototype
     * @param  {string}   collection  collection name
     * @param  {string | null}    itemId  entry id, if null a random id will be assigned to the item
     * @param  {CollectionType}   itemData  ObjectStore data
     * @return {Promise<CollectionEntry<CollectionType>>}
     * @example collection.add("food", "cheese101", {"name": "cheese burger","toppings": "cheese"})
     */
    add(collection: string, itemId: string | null, itemData: CollectionType): Promise<CollectionEntry<CollectionType>>;
    /**
     * Update entry in the collection
     * @link https://getstream.io/docs/collections_introduction/?language=js#updating-collections
     * @method update
     * @memberof Collections.prototype
     * @param  {string}   collection  collection name
     * @param  {string}   entryId  Collection object id
     * @param  {CollectionType}   data  ObjectStore data
     * @return {Promise<CollectionEntry<CollectionType>>}
     * @example store.update("0c7db91c-67f9-11e8-bcd9-fe00a9219401", {"name": "cheese burger","toppings": "cheese"})
     * @example store.update("food", "cheese101", {"name": "cheese burger","toppings": "cheese"})
     */
    update(collection: string, entryId: string, data: CollectionType): Promise<CollectionEntry<CollectionType>>;
    /**
     * Delete entry from collection
     * @link https://getstream.io/docs/collections_introduction/?language=js#removing-collections
     * @method delete
     * @memberof Collections.prototype
     * @param  {string}   collection  collection name
     * @param  {string}   entryId  Collection entry id
     * @return {Promise<APIResponse>} Promise object
     * @example collection.delete("food", "cheese101")
     */
    delete(collection: string, entryId: string): Promise<APIResponse>;
    /**
     * Upsert one or more items within a collection.
     * @link https://getstream.io/docs/collections_batch/?language=js#upsert
     * @method upsert
     * @memberof Collections.prototype
     * @param  {string}   collection  collection name
     * @param {NewCollectionEntry<CollectionType> | NewCollectionEntry<CollectionType>[]} data - A single json object or an array of objects
     * @return {Promise<UpsertCollectionAPIResponse<CollectionType>>}
     */
    upsert(collection: string, data: NewCollectionEntry<CollectionType> | NewCollectionEntry<CollectionType>[]): Promise<UpsertCollectionAPIResponse<CollectionType>>;
    /**
     * Select all objects with ids from the collection.
     * @link https://getstream.io/docs/collections_batch/?language=js#select
     * @method select
     * @memberof Collections.prototype
     * @param {string} collection  collection name
     * @param {string | string[]} ids - A single object id or an array of ids
     * @return {Promise<SelectCollectionAPIResponse<CollectionType>>}
     */
    select(collection: string, ids: string | string[]): Promise<SelectCollectionAPIResponse<CollectionType>>;
    /**
     * Remove all objects by id from the collection.
     * @link https://getstream.io/docs/collections_batch/?language=js#delete_many
     * @method delete
     * @memberof Collections.prototype
     * @param {string} collection  collection name
     * @param {string | string[]} ids - A single object id or an array of ids
     * @return {Promise<APIResponse>}
     */
    deleteMany(collection: string, ids: string | string[]): Promise<APIResponse>;
}
export {};
//# sourceMappingURL=collections.d.ts.map